#include <libpv_config.h>

#include <assert.h>
#include <errno.h>
#include <string.h>

#include <xmlconfig.h>

#include <libpv.h>

#include <pv_bs.h>
#include <pv_bs_8bit.h>
#include <pv_lock.h>
#include <pv_destroy.h>
#include <jv_server.h>


int  _libpv_debuglevel = LEVEL_INFO;


static int pv_init_init_pv(struct pv	*pv,
			   xmlconfig_t	*xmlcfg,
			   unsigned int	nr)
{
	char		buf[1024];
	char		xpath[128];
	unsigned long	tmp;
	int		err;

	/* id */
	snprintf(xpath, sizeof(xpath), "//pv[%i]/@id", nr);
	err = xmlconfig_get_string(xmlcfg, xpath, buf, sizeof(buf), nr);
	if (err <= 0) {
		err = -EINVAL;
		dprintf(LEVEL_CRIT, "(%s) pv %d not found, err %d\n",	
			__FUNCTION__, nr, err);
		goto exit_0;
	}

	buf[sizeof(buf)] = 0;
	pv->id = strdup(buf);	/* FIXME: limit size? strndup? (mkl) */
	if (!pv->id) {
		err = -ENOMEM;
		dprintf(LEVEL_CRIT, "(%s) error: could not get memory for id-buffer\n",
			__FUNCTION__);
		goto exit_0;
	}


	/* type */
	snprintf(xpath, sizeof(xpath), "//pv[%i]/type", nr);
	err = xmlconfig_get_string(xmlcfg, xpath, buf, sizeof(buf), nr);
	if (err <= 0) {
		err = -EINVAL;
		dprintf(LEVEL_CRIT, "(%s) type not found for id \"%s\", err %d\n",	
			__FUNCTION__, pv->id, err);
		goto exit_1;
	}
			
	buf[sizeof(buf)] = 0;

	if      (!strcmp(buf, "uint8_t"))
		pv->type = pv_type_uint8;
	else if (!strcmp(buf, "uint16_t"))
		pv->type = pv_type_uint16;
	else if (!strcmp(buf, "uint32_t"))
		pv->type = pv_type_uint32;

	else if (!strcmp(buf, "int8_t"))
		pv->type = pv_type_int8;
	else if (!strcmp(buf, "int16_t"))
		pv->type = pv_type_int16;
	else if (!strcmp(buf, "int32_t"))
		pv->type = pv_type_int32;

	else if (!strcmp(buf,"string"))
		pv->type = pv_type_string;
	else {
		err = -EINVAL;
		dprintf(LEVEL_CRIT, "(%s) unknown type, \"%s\" for id \"%s\"\n",
			__FUNCTION__, buf, pv->id);
		goto exit_1;
	}


	/* offset */
	snprintf(xpath, sizeof(xpath), "//pv[%i]/offset", nr);
	err = xmlconfig_get_ulong(xmlcfg, xpath, &tmp, 1);	/* HACK: (mkl) */
	if (err <= 0) {
		err = -EINVAL;
		dprintf(LEVEL_CRIT, "(%s) offset not found for id \"%s\", err %d\n",	
			__FUNCTION__, pv->id, err);
		goto exit_1;
	}
	pv->offset = tmp;


	/* name */
	snprintf(xpath, sizeof(xpath), "//pv[%i]/name", nr);
	err = xmlconfig_get_string(xmlcfg, xpath, buf, sizeof(buf), 1);
	if (err <= 0) {
		err = -EINVAL;
		dprintf(LEVEL_CRIT, "(%s) name not found for id \"%s\", err %d\n",	
			__FUNCTION__, pv->id, err);
		goto exit_1;
	}

	buf[sizeof(buf)] = 0;
	pv->name = strdup(buf);	/* FIXME: limit size? strndup? (mkl) */
	if (!pv->name) {
		err = -ENOMEM;
		dprintf(LEVEL_CRIT, "(%s) error: could not get memory for name-buffer\n",
			__FUNCTION__);
		goto exit_1;
	}


	/* size */
	snprintf(xpath, sizeof(xpath), "//pv[%i]/size", nr);
	err = xmlconfig_get_ulong(xmlcfg, xpath, &tmp, 1);	/* HACK: (mkl) */
	if (err == 0) {		/* no match, thus no size :) */
		switch (pv->type) {
		case pv_type_int8:
		case pv_type_uint8:
			pv->size = 1;
			break;
		case pv_type_int16:
		case pv_type_uint16:
			pv->size = 2;
			break;
		case pv_type_int32:
		case pv_type_uint32:
			pv->size = 4;
			break;
		case pv_type_int64:
		case pv_type_uint64:
			pv->size = 8;
			break;
		case pv_type_string:
			err = -EINVAL;
			dprintf(LEVEL_CRIT, "(%s) size not specified for string or undef type,"
				"id \"%s\"\n",
				__FUNCTION__, pv->id);
		case pv_type_undef:
			break;
		}
	} else if (err < 0) {
		dprintf(LEVEL_CRIT, "(%s) size not found for id \"%s\", err %d\n",	
			__FUNCTION__, pv->id, err);
		goto exit_2;
	} else {
		pv->size = tmp;
	}


	/* default */
	snprintf(xpath, sizeof(xpath), "//pv[%i]/@val", nr);
	if (pv->type != pv_type_undef &&
	    pv->type != pv_type_string) {
		err = xmlconfig_get_ulong(xmlcfg, xpath, &tmp, 1);	/* HACK: (mkl) */
		if (err > 0) {
			switch (pv->type) {
			case pv_type_int8:
				pv->dflt.i8 = tmp;
				break;
			case pv_type_uint8:
				pv->dflt.u8 = tmp;
				break;
			case pv_type_int16:
				pv->dflt.i16 = tmp;
				break;
			case pv_type_uint16:
				pv->dflt.u16 = tmp;
				break;
			case pv_type_int32:
				pv->dflt.i32 = tmp;
				break;
			case pv_type_uint32:
				pv->dflt.u32 = tmp;
				break;
			case pv_type_int64:
				pv->dflt.i64 = tmp;
				break;
			case pv_type_uint64:
				pv->dflt.u64 = tmp;
				break;
			case pv_type_string:
			case pv_type_undef:
				break;
			}
		} else {
			pv->dflt.u64 = 0;
		}
	} else if (pv->type == pv_type_string) {
		pv->dflt.string = calloc(1, pv->size);
		if (!pv->dflt.string) {
			err = -ENOMEM;
			dprintf(LEVEL_CRIT, "(%s) error: could not get memory for dflt-buffer for id \"%s\"\n",
				__FUNCTION__, pv->id);
			goto exit_2;
		}
		err = xmlconfig_get_string(xmlcfg, xpath, pv->dflt.string, pv->size, 1);
	}


	/* FIXME: implement stuff for scale, unit, min, max (mkl)*/

	return 0;

 exit_2:
	free(pv->name);
 exit_1:
	free(pv->id);
 exit_0:
	return err;
}


static int pv_init_read_pvs(struct pv_engine	*engine,
			    const char		*xml_filename)
{
	struct pv	pv;
	xmlconfig_t	*xmlcfg;
	int		no_pvs;
	int		err;
	int		ret;
	int		i;
	char		buf;

	/* analyse xml description file */
	xmlcfg = xmlconfig_init(xml_filename);
	if (!xmlcfg) {
		err = -ENOENT;
		dprintf(LEVEL_CRIT, "(%s) could not open XML file \"%s\"\n",
			__FUNCTION__, xml_filename);
		goto exit_0;
	}


	no_pvs = xmlconfig_get_string(xmlcfg, "//pv", &buf, 0, 0);
	if (no_pvs == 0) {
		err = -ENODEV;
		dprintf(LEVEL_WARN, "(%s) no pv entries in XML file found\n",
			__FUNCTION__);
		goto exit_1;
	} else if (no_pvs < 0) {
		err = no_pvs;
		dprintf(LEVEL_CRIT, "(%s) xmlconfig_get_string failed, err %d\n",
			__FUNCTION__, err);
		goto exit_1;
	}

	dprintf(LEVEL_INFO, "(%s) %d pv entries found\n",
		__FUNCTION__, no_pvs);


	engine->pv_list = list_init();
	for (i = 1; i <= no_pvs; i++) {
		err = pv_init_init_pv(&pv, xmlcfg, i);
		if (err)
			goto exit_2;

		engine->pv_list = list_add_tail(engine->pv_list,
						(void *)&pv,
						sizeof(struct pv));
	}

	err = xmlconfig_finish(xmlcfg);
	
	return err;

 exit_2:
	ret = pv_destroy_free_pvs(engine->pv_list);
	if (ret)
		dprintf(LEVEL_ERROR, "(%s) failed to destroy pv_list, err %d\n",
			__FUNCTION__, ret);
 exit_1:
	ret = xmlconfig_finish(xmlcfg);
	if (ret)
		dprintf(LEVEL_ERROR, "(%s) xmlconfig_finish failed, err %d\n",
			__FUNCTION__, ret);
 exit_0:
	return err;
}


static void pv_init_show_pvs(struct pv_engine	*engine)
{
	struct pv	*pv;
	int		finished = 0; 

	printf("process variables:\n");
	printf("------------------\n\n");
		
	for (goto_head(&engine->pv_list); !finished; goto_next(&engine->pv_list)) {
		pv = (struct pv *)list_get_element_data(engine->pv_list);
			
		if (is_last(engine->pv_list))
			finished=1;

		printf("id \"%s\", offset %d\n",
		       pv->id, (int)pv->offset);
	}

	printf("\n");
}


static size_t pv_init_get_pvs_size(ptxlist	*pv_list)
{
	size_t	size = 0;
	int	finished = 0;

	for (goto_head(&pv_list); !finished; goto_next(&pv_list)) {
		struct pv *pv = (struct pv *)list_get_element_data(pv_list);
			
		if (is_last(pv_list))
			finished=1;
			
		size = MAX(size, pv->offset + pv->size);
	}

	dprintf(LEVEL_INFO, "(%s) store size %zd\n",
		__FUNCTION__, size);

	return size;
}


int pv_init(pv_engine_t		*engine_anon,
	    pv_engine_attr_t	*attr)
{
	struct pv_engine	*engine = (struct pv_engine *)engine_anon;
	int			err;
	size_t			size;

	if (!engine || !attr)
		return -EINVAL;

	if (attr->xml_filename[0] == 0 ||
	    attr->store_filename[0] == 0)
		return -EINVAL;

	memset(engine, 0, sizeof(struct pv_engine));

	/* read pv from xml into list */
	err = pv_init_read_pvs(engine, attr->xml_filename);
	if (err)
		goto exit_0;

	/* show the pvs we've found */
	if (_libpv_debuglevel >= LEVEL_DEBUG)
		pv_init_show_pvs(engine);

	
	/* find out size of pv area */
	size = pv_init_get_pvs_size(engine->pv_list);
	assert(size != 0);
	
	
	/* init locking */
	err = pv_lock_init(engine, attr->store_filename);
	if (err)
		goto exit_1;


	/* init the backing store */
	err = pv_bs_init(engine, attr->store_type, attr->store_filename, size);
	if (err)
		goto exit_1;	/* FIXME: destroy locking (mkl) */

	return 0;

 exit_1:
	pv_destroy_free_pvs(engine->pv_list);
 exit_0:
	return err;
}
